import sys
from os          import listdir
from rdflib      import Namespace
from scry.config import get_conf, get_subdir, write_conf

BLAST           = Namespace('http://scry.rocks/blast/')

CONF = get_conf()
if 'scry_blast' not in CONF.sections():
    # Set up the default configuration
    CONF.add_section('scry_blast')
    CONF.set('scry_blast','bin_dir','UNKNOWN')
    CONF.set('scry_blast','db_dir', get_subdir('Services','BLAST','db'))
    CONF.set('scry_blast','default_db', 'UNKNOWN')
    CONF.set('scry_blast','cache_seqs','True')
    CONF.set('scry_blast','cache_dir', get_subdir('Services','BLAST','sequence_cache'))
    write_conf(CONF)
    print "\n  New configuration parameters added to SCRY's config.ini for the `scry_blast` service."
    print "  Remember to configure `bin_dir` and add at least one database to the appropriate directory"
    print "  before importing this service with SCRY!\n"
    sys.exit()

DATABASES = list()
for f in listdir(CONF.get('scry_blast','db_dir')):
    if f.endswith('.psq'):
        DATABASES.append(f[:-4])

DEFAULT_DB = CONF.get('scry_blast','default_db')
if DEFAULT_DB == 'UNKNOWN':
    DEFAULT_DB = DATABASES[0]


# The SCRY BLAST scripts are intended for use with NCBI BLAST software version 2.2.29+
# Currently supported programs are blastn, blastp, blastx, tblastn and tblastx.

__version__     = '0.2'

import options

if False:       # Have the binaries?
    pass
elif False:     # Have at least one database?
    pass
else:           # Import the procedures
    from run_blast      import RunBLAST
    from fetch_sequence import FetchSequence



from scry.services.classes import Procedure, Argument
from scry.utility          import SCRYError

from scry_blast            import __version__
from scry_blast.config     import *
from scry_blast.options    import *

from rdflib.term           import Literal

from os                    import system
from os.path               import join
#from subprocess            import check_call
from xml.etree.ElementTree import parse

def run_blast(inputs,outputs,handler):
    
    def make_files():
        temp_dir = handler.get_temp_dir()
        out_file = join(temp_dir,'output.xml')
        if not ('seq' in inputs) ^ ('file' in inputs):
            raise SCRYError("SCRY BLAST requires exactly one of 'seq' or 'file' to be specified.")
        elif 'seq' in inputs:
            seq = inputs['seq'].encode()
            if not seq.startswith('>'):
                seq = '>query_sequence\n%s' % seq
            in_file = join(temp_dir,'input.fasta')
            with open(in_file,'w') as f: f.write(seq)
        else:
            in_file = inputs['file'].encode()
        return in_file, out_file
        
    def get_program():
        try:
            return inputs['program'].encode()
        except KeyError:
            if 'seq_type' in inputs:
                seq_type = inputs['seq_type'].encode().lower()
            else:
                seq_type = guess_type()
            if seq_type == 'protein':
                return 'blastp'
            else:
                return 'blastn'
    
    def guess_type():
        with open(in_file) as f:
            f.readline() # Skip the first line
            alphabet  = set(f.read().lower())
            prot_only = set('efijlopqz') # Characters only used for amino acid codes in the FASTA format
            if len(alphabet) > 12 or alphabet.intersection(prot_only):
                # If more than 12 different characters are used, or if any characters unique to proteins are used...
                return 'protein'
            else:
                return 'nucleotide'

    def get_cmd():
        cmd = '%s -query %s -out %s -db %s -outfmt 5' % (program_file, in_file, out_file, db_file)
        for k in inputs:
            if k in ALLOWED_OPTIONS and ALLOWED_OPTIONS[k]:
                cmd += ' -%s %s' % (k, inputs[k])
            elif k in ALLOWED_FLAGS and ALLOWED_FLAGS[k]:
                cmd += ' -%s' % k
        return cmd

    def parse_XML():
        hsp_dicts = list()
        with open(out_file) as f:
            doc = parse(f)
        root = doc.getroot()
        hits = root[-1][-1][-2].getchildren() # Output_iterations / Last iteration / Iteration hits
        for hit in hits:
            hit_dict = {elem.tag:elem.text for elem in hit.getchildren()}
            hsps = hit[-1].getchildren()
            for hsp in hsps:
                hsp_dict = {elem.tag:elem.text for elem in hsp.getchildren()}
                hsp_dict.update(hit_dict)
                for attr in ['identity','positive','gaps']:
                    frac = float(hsp_dict['Hsp_%s' % attr]) / float(hsp_dict['Hsp_align-len'])
                    hsp_dict['frac_%s' % attr] = frac
                    hsp_dict['perc_%s' % attr] = frac * 100
                hsp_dicts.append(hsp_dict)
        return hsp_dicts


    in_file, out_file = make_files()
    program           = get_program()
    program_file      = join(CONF.get('scry_blast','bin_dir'),program)
    db                = (inputs['db'].encode() if 'db' in inputs else DEFAULT_DB)
    db_file           = join(CONF.get('scry_blast','db_dir'),db)
    cmd               = get_cmd()

    system(cmd)
    hits     = parse_XML()
    num_hits = {}
    if 'num_hits' in outputs:
        num_hits = {'num_hits':Literal(len(hits))}
        outputs.remove('num_hits')
    
    out = list()
    for d in hits:
        od = {k : Literal(d[out_dict[k][0]]) for k in outputs}
        od.update(num_hits)
        out.append(od)
    
    return out

# PROCEDURE
p = Procedure(BLAST.blast)
p.function    = run_blast
p.author      = "Bas Stringer"
p.description = "The SCRY BLAST procedure; a bridge between NCBI's BLAST program suite and SPARQL"
p.provenance  = "Results generated by SCRY BLAST version %s" % __version__
p.version     = __version__
RunBLAST      = p

# ARGUMENTS
in_dict = {'seq'      : "A FASTA formatted input sequence BLAST should query with\nEither this or the 'file' argument *must* be specified",
           'seq_type' : "'protein' or 'nucleotide'",
           'file'     : "A URIRef with a file path to the input sequence BLAST should query with\nEither this or the 'seq' argument *must* be specified",
           'program'  : "The BLAST program you wish to run\nMust be one of; %s" % ', '.join(SPECIAL_OPTIONS['program']),
           'db'       : "The database BLAST should query\nThis Orb offers access to the following databases;\n%s" % ', '.join(SPECIAL_OPTIONS['db'])}

for k in in_dict:
    a             = Argument(k)
    a.uri         = BLAST['input/%s' % k]
    a.description = in_dict[k]
    RunBLAST.add_input(a,k=='seq')

options = dict()
options.update(ALLOWED_OPTIONS)
options.update(ALLOWED_FLAGS)
for opt in options:
    if not options[opt]: continue
    a = Argument(opt)
    a.uri = BLAST['option/%s' % opt]
    a.description = "SCRY BLAST option '%s'\nRefer to NCBI's BLAST documentation for more details" % opt
    a.valuetype   = Literal
    RunBLAST.add_input(a)

out_dict = {'num_hits'      : ( None             , "The total number of high-scoring segment pairs BLAST found in the database"),
            'id'            : ('Hit_id'          , "The identifier line of a query hit"),
            'def'           : ('Hit_def'         , "The definition line of a query hit"),
            'ac'            : ('Hit_accession'   , "The accession number of a query hit"),
            'len'           : ('Hit_len'         , "The length of a hit sequence"),
            'bitscore'      : ('Hsp_bit-score'   , "The bit-score of a hit; a log-scaled variant of the 'score' output"),
            'score'         : ('Hsp_score'       , "The raw score of a hit"),
            'e_val'         : ('Hsp_evalue'      , "The Expectation value of a hit"),
            'qry_from'      : ('Hsp_query-from'  , "The coordinate on the query sequence where this hit's alignment starts"),
            'qry_to'        : ('Hsp_query-to'    , "The coordinate on the query sequence where this hit's alignment ends"),
            'qry_frame'     : ('Hsp_query-frame' , "The reading frame for which the query sequence was translated;\nUsed by blastx, tblastx and tblastn"),
            'hit_from'      : ('Hsp_hit-from'    , "The coordinate on the hit sequence where this hit's alignment starts"),
            'hit_to'        : ('Hsp_hit-to'      , "The coordinate on the hit sequence where this hit's alignment ends"),
            'hit_frame'     : ('Hsp_hit-frame'   , "The reading frame for which the hit sequence was translated;\nUsed by blastx, tblastx and tblastn"),
            'identity'      : ('Hsp_identity'    , "The number of identical pairs of residues in the hit's alignment"),
            'positive'      : ('Hsp_positive'    , "The number of pairs of residues in the hit's alignment with a positive score given the used substitution matrix"),
            'gaps'          : ('Hsp_gaps'        , "The number of gaps in the alignment between the query and hit sequence"),
            'frac_identity' : ('frac_identity'   , "The fraction of identical pairs of residues in the hit's alignment"),
            'frac_positive' : ('frac_positive'   , "The fraction of aligned residues in the hit's alignment with a positive score given the used substitution matrix"),
            'frac_gaps'     : ('frac_gaps'       , "The fraction of gaps in the alignment between the query and hit sequence"),
            'perc_identity' : ('perc_identity'   , "The percentage of identical pairs of residues in the hit's alignment"),
            'perc_positive' : ('perc_positive'   , "The percentage of pairs of residues in the hit's alignment with a positive score given the used substitution matrix"),
            'perc_gaps'     : ('perc_gaps'       , "The percentage of gaps in the alignment between the query and hit sequence"),
            'align_len'     : ('Hsp_align-len'   , "The number of residues in the high-scoring segment pair"),
            'qry_seq'       : ('Hsp_qseq'        , "The aligned section of the query sequence"),
            'hit_seq'       : ('Hsp_hseq'        , "The aligned section of the hit sequence"),
            'midline'       : ('Hsp_midline'     , "The midline between 'qry_seq' and 'hit_seq'")}

for k in out_dict:
    a             = Argument(k)
    a.uri         = BLAST['output/%s' % k]
    a.description = out_dict[k][1]
    RunBLAST.add_output(a,k=='seq')


import re
import pkg_resources as pkr

from rdflib.term           import URIRef, Literal
from urllib2               import urlopen
from os.path               import join, isfile

from scry.services.classes import Procedure, Argument
from scry.utility          import SCRYError, make_bool

from scry_blast            import __version__
from scry_blast.config     import BLAST, CONF

CACHE_SEQUENCES = make_bool(CONF.get('scry_blast','cache_seqs'))
CACHE_DIRECTORY = CONF.get('scry_blast','cache_dir')
if CACHE_SEQUENCES:
    pkr.ensure_directory(CACHE_DIRECTORY)

def fetch_sequence(inputs,outputs,handler):
    if 'id' not in inputs:
        raise SCRYError("The FetchSequence procedure requires an identifier to be specified as input argument.")
    use_cache  = (True if CACHE_SEQUENCES and not 'reload' in inputs else False)
    identifier = inputs['id'].encode()
    source     = (inputs['source'].encode() if 'source' in inputs and inputs['source'].lower() in KNOWN_SOURCES else DEFAULT_SOURCE).lower()
    id_split   = re.split('\W',identifier)
    try:
        while not re.match(ID_EXPRESSIONS[source],identifier):
            identifier = id_split.pop()
    except IndexError:
        raise SCRYError('No valid %s identifier could be parsed from %s' % (source, inputs['id'].encode()))

    seq_type   = (inputs['seq_type'].encode() if 'seq_type' in inputs else 'protein').lower()
    format     = 'fasta'

    if use_cache or CACHE_SEQUENCES:
        file_dir  = join(CACHE_DIRECTORY,source,seq_type)
    else:
        file_dir  = handler.get_temp_dir()
    file_path = join(file_dir,'%s.%s' % (identifier,format))
    
    if not use_cache or not isfile(file_path):
        pkr.ensure_directory(file_path)
        fetch_fnc = KNOWN_SOURCES[source]
        fetch     = fetch_fnc(identifier,seq_type,inputs)
        with open(file_path,'w') as f:
            f.write(fetch)

    out = dict()
    out['file'] = URIRef(file_path)
    if any([spec in outputs for spec in ['seq','seq_len','definition']]):
        seq, definition   = parse_fasta(file_path)
        out['seq']        = Literal(seq)
        out['seq_len']    = Literal(len(seq))
        out['definition'] = definition
    return out


def fetch_uniprot(identifier,seq_type,inputs):
    # format:   fasta
    # seq_type: protein
    url   = 'http://www.uniprot.org/uniprot/%s.fasta' % identifier
    fetch = urlopen(url).read()
    return fetch


# WARNING: without appening &multiple_sequences=1 to the URL, this function is highly likely to
# fail fetching sequences because of how Ensembl's identifiers work... USE WITH CAUTION!
def fetch_ensembl(identifier,seq_type,inputs):
    # format:   fasta
    # seq_type: genomic,cds,cdna,protein
    url   = 'http://rest.ensembl.org/sequence/id/%s.fasta?type=%s' % (identifier, seq_type)
    fetch = urlopen(url).read()
    return fetch


# This function only parses the first sequence from a file!
def parse_fasta(file_path):
    sequence   = str()
    definition = None
    with open(file_path) as f:
        for line in f:
            line = line.strip()
            if line.startswith('>'):
                if not definition:
                    definition = line
                else:
                    break
            else:
                sequence += line
    return sequence, definition

DEFAULT_SOURCE = 'uniprot'
KNOWN_SOURCES  = {'uniprot' : fetch_uniprot,
                  'ensembl' : fetch_ensembl}
ID_EXPRESSIONS = {'uniprot' : re.compile('^[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}$'),
                  'ensembl' : re.compile('^ENS[A-Z]+[0-9]{11}|[A-Z]{3}[0-9]{3}[A-Za-z](-[A-Za-z])?|CG[0-9]+|[A-Z0-9]+\.[0-9]+|YM[A-Z][0-9]{3}[a-z][0-9]$')}

p = Procedure(BLAST.fetch_sequence)
p.function    = fetch_sequence
p.author      = "Bas Stringer"
p.description = "A function which can fetch biological sequences from a variety of sources\nCurrently supported; %s" % ', '.join(KNOWN_SOURCES.keys())
p.provenance  = "Sequence fetched by SCRY BLAST version %s" % __version__
p.version     = __version__
FetchSequence = p

                                    
arg_dict = {'id'         : ('in'  , "The (source-specific) identifier for which to fetch a sequence"),
            'source'     : ('in'  , "The source to fetch a sequence from\nCurrently supported; %s" % ', '.join(KNOWN_SOURCES.keys())),
            'format'     : ('in'  , "The format in which to fetch a sequence\nCurrently supported; fasta"),
            'seq_type'   : ('in'  , "The type of sequence to fetch; Currently supported;\nUniProt; protein\nEnsembl; genomic, cdna, cds, protein"),
            'seq_db'     : ('in'  , "Used to narrow down which database to retrieve sequences from, where applicable.\nCurrently used by; ensembl"),
            'reload'     : ('in'  , "A boolean, used to overwrite a cached file if one exists.\nIgnored if caching is turned off or if no file is cached for the requested id."),
            'seq'        : ('out' , "The sequence string that was fetched"),
            'seq_len'    : ('out' , "The length of the fetched sequence"),
            'definition' : ('out' , "The definition or accession string of the fetched sequence"),
            'file'       : ('out' , "The path to the file where the fetched sequence is stored%s" % ('' if CACHE_SEQUENCES else ' temporarily'))}

for key in arg_dict:
    a             = Argument(key)
    a.uri         = BLAST['fetch/%s' % key]
    a.description = arg_dict[key][1]
    a.valuetype   = Literal
    if arg_dict[key][0] == 'in':
        if key == 'id':
            FetchSequence.add_input(a,True,True)
        else:
            FetchSequence.add_input(a)
    else:
        if key == 'file': a.valuetype = URIRef
        FetchSequence.add_output(a,key == 'seq')


from scry_blast.config import DATABASES

SPECIAL_OPTIONS = {'db'                     : DATABASES,
                   'program'                : ['blastn','blastp','blastx','tblastn','tblastx']}

ALLOWED_OPTIONS = {'best_hit_overhang'      : True ,
                   'best_hit_score_edge'    : True ,
                   'comp_based_stats'       : True ,
                   'culling_limit'          : True ,
                   'db_gencode'             : True ,
                   'db_hard_mask'           : False,
                   'db_soft_mask'           : False,
                   'dbsize'                 : False,
                   'dust'                   : True ,
                   'entrez_query'           : False,
                   'evalue'                 : True ,
                   'export_search_strategy' : False,
                   'filtering_db'           : False,
                   'gapextend'              : True ,
                   'gapopen'                : True ,
                   'gilist'                 : False,
                   'import_search_strategy' : False,
                   'in_pssm'                : False,
                   'index_name'             : False,
                   'matrix'                 : True ,
                   'max_hsps'               : True ,
                   'max_intron_length'      : True ,
                   'max_target_seqs'        : True ,
                   'min_raw_gapped_score'   : True ,
                   'negative_gilist'        : False,
                   'num_alignments'         : True ,
                   'num_descriptions'       : True ,
                   'num_threads'            : False,
                   'off_diagonal_range'     : True ,
                   'out'                    : False,
                   'outfmt'                 : False,
                   'penalty'                : True ,
                   'query'                  : False,
                   'query_gencode'          : True ,
                   'query_loc'              : True ,
                   'reward'                 : True ,
                   'searchsp'               : True ,
                   'seg'                    : True ,
                   'seqidlist'              : False,
                   'show_gis'               : False,
                   'soft_masking'           : False,
                   'strand'                 : True ,
                   'subject'                : False,
                   'subject_loc'            : False,
                   'task'                   : False,
                   'template_length'        : False,
                   'template_type'          : False,
                   'threshold'              : True ,
                   'use_index'              : False,
                   'windows_masker_db'      : False,
                   'window_masker_taxid'    : False,
                   'window_size'            : True ,
                   'word_size'              : True ,
                   'xdrop_gap'              : True ,
                   'xdrop_gap_final'        : True ,
                   'xdrop_ungap'            : True }

ALLOWED_FLAGS   = {'h'                      : False,
                   'help'                   : False,
                   'html'                   : False,
                   'lcase_masking'          : True ,
                   'no_greedy'              : True ,
                   'parse_deflines'         : False,
                   'remote'                 : False,
                   'show_gis'               : False,
                   'sum_statistics'         : True ,
                   'ungapped'               : True ,
                   'use_sw_tback'           : True ,
                   'version'                : False}


from setuptools import setup

setup(
    name         = 'scry_blast',
    version      = '0.6',
    description  = "A SCRY service to extend SPARQL with procedures that run NCBI's BLAST program",
    author       = 'Bas Stringer',
    author_email = 'b.stringer@vu.nl',
    url          = 'http://www.few.vu.nl/',
    license      = 'MIT',
    keywords     = 'scry sparql math service extension ncbi blast bioinformatics',
    classifiers  = ['Development Status :: 4 - Beta',
                    'Intended Audience :: Developers',
                    'Intended Audience :: Healthcare Industry',
                    'Intended Audience :: Information Technology',
                    'Intended Audience :: Science/Research',
                    'License :: OSI Approved :: MIT License',
                    'Operating System :: OS Independent',
                    'Programming Language :: Python',
                    'Programming Language :: Python :: 2.7',
                    'Topic :: Scientific/Engineering :: Bio-Informatics'],

    install_requires = ['scry>=0.1'],
    packages = ['scry_blast']
)
